@@ -3,10 +3,12 @@
 
 """MosaicML Streaming Datasets for cloud-native model training."""
 
-from streaming.base.dataset import Dataset
+from streaming.base.dataloader import StreamingDataLoader
+from streaming.base.dataset import StreamingDataset
 from streaming.base.format import CSVWriter, JSONWriter, MDSWriter, TSVWriter, XSVWriter
 from streaming.base.local import LocalDataset
 
 __all__ = [
-    'Dataset', 'CSVWriter', 'JSONWriter', 'LocalDataset', 'MDSWriter', 'TSVWriter', 'XSVWriter'
+    'StreamingDataLoader', 'StreamingDataset', 'CSVWriter', 'JSONWriter', 'LocalDataset',
+    'MDSWriter', 'TSVWriter', 'XSVWriter'
 ]
