@@ -5,31 +5,27 @@
 
 import json
 import os
-import sys
 from enum import IntEnum
 from multiprocessing.shared_memory import SharedMemory
 from threading import Thread
-from time import sleep, time
+from time import sleep
 from typing import Any, Dict, Iterator, Optional, Tuple
 
 import numpy as np
-import torch
 from filelock import FileLock
 from numpy.typing import NDArray
-from torch import distributed as tdist
 from torch.utils.data import IterableDataset
 
-from streaming.base import distributed as dist
 from streaming.base.compression import decompress
 from streaming.base.format import reader_from_json
 from streaming.base.format.base.reader import FileInfo
 from streaming.base.hashing import get_hash
 from streaming.base.index import Index, get_index_basename
 from streaming.base.partitioning import get_partitions
-from streaming.base.shared import SharedBarrier
+from streaming.base.shared import SharedBarrier, create_shared_memory
 from streaming.base.shuffle import get_shuffle
 from streaming.base.storage import download
-from streaming.base.util import set_mp_start_method
+from streaming.base.util import wait_for_file_to_exist, wait_for_local_leader
 from streaming.base.world import World
 
 # Time to wait, in seconds.
@@ -123,8 +119,7 @@ class StreamingDataset(IterableDataset):
             an exception. Defaults to ``60``.
         validate_hash (str, optional): Optional hash or checksum algorithm to use to validate
             shards. Defaults to ``None``.
-        shuffle_seed (int, optional): Seed for shuffling, or ``None`` for random seed. Defaults to
-            ``None``.
+        shuffle_seed (int): Seed for Deterministic data shuffling. Defaults to ``9176``.
         num_canonical_nodes (int, optional): Canonical number of nodes for shuffling with resumption.
             Defaults to ``None``, which is interpreted as the number of nodes of the initial run.
         batch_size (int, optional): Batch size of its DataLoader, which affects how the dataset is
@@ -141,7 +136,7 @@ class StreamingDataset(IterableDataset):
                  download_retry: int = 2,
                  download_timeout: float = 60,
                  validate_hash: Optional[str] = None,
-                 shuffle_seed: Optional[int] = None,
+                 shuffle_seed: int = 9176,
                  num_canonical_nodes: Optional[int] = None,
                  batch_size: Optional[int] = None):
         self.local = local
@@ -153,14 +148,28 @@ class StreamingDataset(IterableDataset):
         self.download_retry = download_retry
         self.download_timeout = download_timeout
         self.validate_hash = validate_hash or None
-        set_mp_start_method(sys.platform)
 
-        if tdist.is_available() and not tdist.is_initialized() and torch.cuda.is_available() and \
-                'RANK' in os.environ:
-            tdist.init_process_group('nccl')
+        if self.download_retry < 0:
+            raise ValueError('Parameter ``download_retry`` must be non-negative')
+        if self.download_timeout < 0:
+            raise ValueError(
+                'Parameter ``download_timeout`` (in seconds) must be greater than zero')
+
+        # Placeholder for _resume_shm, a shared memory object where load_state_dict() saves its
+        # data to be picked up by __iter__().
+        self._resume_shm = None
+
+        # Partition state.
+        self._partition_state = None
+
+        # Initialize the World context.
+        #
+        # Beware: This information is for the per-rank process. DataLoader worker processes may see
+        # different values for these fields. We are saving the rank World here because we cannot
+        # instantiate a World inside the StreamingDataset destructor.
+        self._rank_world = world = World()
 
         # Seed is set below.
-        world = World()
         self.num_canonical_nodes = num_canonical_nodes
         self.batch_size = batch_size
         self.shuffle_seed = shuffle_seed
@@ -171,7 +180,11 @@ class StreamingDataset(IterableDataset):
             filename = self._download_file(basename)
         else:
             filename = os.path.join(local, self.split, basename)  # pyright: ignore
-        dist.barrier()
+
+        # Everyone waits for the file to become populated.
+        wait_for_file_to_exist(filename, TICK, self.download_timeout,
+                               f'{filename} file took too long to download')
+
         obj = json.load(open(filename))
         if obj['version'] != 2:
             raise ValueError('Unsupported version')
@@ -187,57 +200,41 @@ class StreamingDataset(IterableDataset):
         self.index = Index(self.shard_sizes)
 
         # Determine and distribute shuffle seed and shm prefix.
-        if shuffle_seed is None:
-            shuffle_seed = np.random.randint(1 << 60)
+        seed_rng = np.random.default_rng(shuffle_seed)
+        self.shuffle_seed = int(seed_rng.integers(1 << 60))
         prefix_int = np.random.randint(1 << 24)
-        if world.num_ranks > 1:
-            # Setup for coordinating.
-            device_prefix = 'cuda' if torch.cuda.is_available() else 'cpu'
-            device = torch.device(f'{device_prefix}:{world.rank_of_node}')
-            tensor = torch.zeros(1, dtype=torch.int64, device=device)
-
-            # Coordinate the shuffle seed across ranks.
-            if world.is_leader:
-                tensor[0] = shuffle_seed
-            dist.broadcast(tensor, 0)
-            shuffle_seed = int(tensor)
-
-            # Add a coordinated random prefix to all shm names for uniqueness.
-            if world.is_leader:
-                tensor[0] = prefix_int
-            dist.broadcast(tensor, 0)
-            prefix_int = int(tensor)
-        self.shuffle_seed = shuffle_seed
         self._prefix = f'{prefix_int:06x}'
 
+        # Should be a unique shared directory per each StreamingDataset instantiation to avoid a conflict
+        # between a different StreamingDataset instance on a same machine.
+        self._shared_dir = os.path.join(os.path.sep, 'tmp', 'streaming', self._prefix)
+
+        # Create the shared memory-backed worker barrier, without its lock, which is unpickleable.
+        worker_barrier_filelock_path = os.path.join(self._shared_dir, 'barrier_filelock')
+        worker_barrier_shm_path = f'{self._prefix}_barrier'
+        self._worker_barrier = SharedBarrier(worker_barrier_filelock_path, worker_barrier_shm_path,
+                                             world.is_local_leader)
+
+        # Remove the lock that makes it unpickleable
+        del self._worker_barrier.lock
+
         # Set up the epoch counter.
         #
         # Note: we do not assume that the end of __iter__() will ever be reached, so we need to
         # increment the epoch counter at the start of __iter__() instead of at the end, so we need
         # to track what the next epoch is, not the current epoch.
-        name = f'{self._prefix}_next_epoch'
-        size = np.int64().nbytes
-        try:
-            self._next_epoch_shm = SharedMemory(name, True, size)
-        except FileExistsError:
-            sleep(TICK)
-            self._next_epoch_shm = SharedMemory(name, False, size)
+        self._next_epoch_shm = create_shared_memory(name=f'{self._prefix}_next_epoch',
+                                                    size=np.int64().nbytes)
         self._next_epoch_arr = np.ndarray(1, buffer=self._next_epoch_shm.buf, dtype=np.int64)
         self._next_epoch_arr[0] = 0
 
-        # Placeholder for _resume_shm, a shared memory object where load_state_dict() saves its
-        # data to be picked up by __iter__().
-        self._resume_shm = None
+        # Get the filelock filename that protects shard_states shared memory array.
+        self.shard_states_filename = os.path.join(self._shared_dir, '_shard_states_filelock')
 
-        # Create the barrier.
-        self._worker_barrier_filelock_path = os.path.join(os.path.sep, 'tmp', 'streaming',
-                                                          self._prefix, 'barrier_filelock')
-        self._worker_barrier_shm_path = f'{self._prefix}_barrier'
-        self._worker_barrier = SharedBarrier(self._worker_barrier_filelock_path,
-                                             self._worker_barrier_shm_path)
-
-        # Partition state.
-        self._partition_state = None
+        # Create or attach shard_states array (tells if each shard is unknown, downloading, or
+        # downloaded).
+        self._shard_states = create_shared_memory(name=f'{self._prefix}_shard_states',
+                                                  size=len(self.shard_sizes) * np.uint8(0).nbytes)
 
     @property
     def next_epoch(self) -> int:
@@ -314,10 +311,6 @@ class StreamingDataset(IterableDataset):
 
         # Check if the resume state is stale.
         if obj['epoch'] < epoch:
-            # Clean up stale state.
-            if world.is_local_leader:
-                shm.close()
-                shm.unlink()
             self._set_canonical_num_nodes(world)
             return epoch, 0
 
@@ -338,6 +331,9 @@ class StreamingDataset(IterableDataset):
         Returns:
             Tuple[int, int]: What epoch this is, and sample offset in that epoch.
         """
+        # Reference the same shared memory object in a worker process
+        self._next_epoch_arr = np.ndarray(1, buffer=self._next_epoch_shm.buf, dtype=np.int64)
+
         # Either resume from checkpoint, or start from scratch.
         presumed_epoch = self.next_epoch
         epoch, sample_in_epoch = self._resume(world, presumed_epoch)
@@ -355,7 +351,7 @@ class StreamingDataset(IterableDataset):
                        world: World,
                        epoch: int,
                        sample_in_epoch: int,
-                       timeout: Optional[float] = 60) -> NDArray[np.int64]:
+                       timeout: float = 60) -> NDArray[np.int64]:
         """Get this worker's partition of this epoch's sample space.
 
         Args:
@@ -375,9 +371,8 @@ class StreamingDataset(IterableDataset):
             raise RuntimeError('Shuffle seed can never be None')
 
         # Decide where to save shuffle data.
-        tmp_filename = os.path.join(os.path.sep, 'tmp', 'streaming', self._prefix,
-                                    'shuffle.npy.tmp')
-        filename = os.path.join(os.path.sep, 'tmp', 'streaming', self._prefix, 'shuffle.npy')
+        tmp_filename = os.path.join(self._shared_dir, 'shuffle.npy.tmp')
+        filename = os.path.join(self._shared_dir, 'shuffle.npy')
 
         # In the local leader, generate this epoch's global sample ordering, then save to file.
         # Tensor shape: (num nodes, ranks per node, workers per rank, samples per worker).
@@ -394,17 +389,7 @@ class StreamingDataset(IterableDataset):
             os.rename(tmp_filename, filename)
 
         # Everyone waits for the file to become populated.
-        t0 = time()
-        while True:
-            sleep(TICK)
-            if os.path.exists(filename):
-                sleep(TICK)
-                break
-            if timeout is not None:
-                dt = time() - t0
-                if timeout < dt:
-                    raise RuntimeError('Partitioning and shuffling took too long, bailing out: ' +
-                                       f'{timeout:.3f} < {dt:.3f} sec.')
+        wait_for_file_to_exist(filename, TICK, timeout, 'Partitioning and shuffling took too long')
 
         # Each worker loads its slice of the sample ID tensor to iterate through.
         # Tensor shape: (num nodes, ranks per node, workers per rank, samples per worker).
@@ -579,32 +564,20 @@ class StreamingDataset(IterableDataset):
         else:
             raise RuntimeError('Unknown shard state')
 
-    def _get_shard_states(self) -> Tuple[FileLock, NDArray[np.uint8], SharedMemory]:
+    def _get_shard_states(self) -> Tuple[FileLock, NDArray[np.uint8]]:
         """Get the shared shard states array and its protecting lock.
 
-        Also returns the shared memory object to keep a reference around, preventing garbage
-        collection.
-
         Returns:
-            Tuple[FileLock, NDArray[np.uint8], SharedMemory]: Lock, array, and shared memory.
+            Tuple[FileLock, NDArray[np.uint8]]: Lock, and array.
         """
         # Get the filelock that protects shard_states shared memory array.
-        filename = os.path.join(os.path.sep, 'tmp', 'streaming', self._prefix,
-                                '_shard_states_filelock')
-        lock = FileLock(filename)
+        lock = FileLock(self.shard_states_filename)
 
-        # Create or attach shard_states array (tells if each shard is unknown, downlaoding, or
-        # downloaded).
-        name = f'{self._prefix}_shard_states'
-        size = len(self.shard_sizes) * np.uint8(0).nbytes
-        try:
-            shm = SharedMemory(name, True, size)
-        except FileExistsError:
-            sleep(TICK)
-            shm = SharedMemory(name, False, size)
-        shard_states = np.ndarray(len(self.shard_sizes), buffer=shm.buf, dtype=np.uint8)
+        shard_states = np.ndarray(len(self.shard_sizes),
+                                  buffer=self._shard_states.buf,
+                                  dtype=np.uint8)
 
-        return lock, shard_states, shm
+        return lock, shard_states
 
     def _download_thread(self, state: _PartitionState) -> None:
         """Download the relevant shards in the background while we are being iterated.
@@ -618,7 +591,7 @@ class StreamingDataset(IterableDataset):
         Args:
             state (_PartitionState): The partition state.
         """
-        shard_states_lock, shard_states, shm = self._get_shard_states()
+        shard_states_lock, shard_states = self._get_shard_states()
 
         # Download loop.
         while True:
@@ -650,8 +623,6 @@ class StreamingDataset(IterableDataset):
             self._download_or_skip_shard(shard_states_lock, shard_states, shard_id)
             state.download_index += 1
 
-        del shm
-
     def _ready_thread(self, state: _PartitionState) -> None:
         """Download the relevant shards in the background while we are being iterated.
 
@@ -664,7 +635,7 @@ class StreamingDataset(IterableDataset):
         Args:
             state (_PartitionState): The partition state.
         """
-        _, shard_states, shm = self._get_shard_states()
+        _, shard_states = self._get_shard_states()
 
         # Download loop.
         while True:
@@ -697,8 +668,6 @@ class StreamingDataset(IterableDataset):
                 sleep(TICK)
             state.ready_index += 1
 
-        del shm
-
     def _each_sample(self, sample_ids: NDArray[np.int64]) -> Iterator[int]:
         """Iterate over each sample ID, while downloading ahead in the background.
 
@@ -719,6 +688,11 @@ class StreamingDataset(IterableDataset):
         Returns:
             Iterator[Dict[str, Any]]: Each sample.
         """
+        # Lazily create the worker barrier's FileLock, which contains a threading Lock, which is
+        # unpickleable.
+        if not hasattr(self._worker_barrier, 'lock'):
+            self._worker_barrier.lock = FileLock(self._worker_barrier.filelock_path)
+
         # Exit the thread that is downloading the shards for last epoch, if it exists.
         if self._partition_state:
             self._partition_state.stop()
@@ -791,3 +765,33 @@ class StreamingDataset(IterableDataset):
             sleep(TICK)
             self._resume_shm = SharedMemory(name)
             assert len(self._resume_shm.buf) == len(data)
+
+    def _cleanup_shared_memory(self, shm: Any, world: World) -> None:
+        """Clean up the shared memory resources.
+
+        Args:
+            shm (Any): A SharedMemory object
+            world (World): World state.
+        """
+        if shm is not None:
+            # Close each SharedMemory instance
+            shm.close()
+            if world.is_local_leader:
+                # Call unlink only once to release the shared memory
+                shm.unlink()
+            else:
+                # Wait for local leader process to execute first
+                sleep(1)
+
+    def __del__(self):
+        # Wait for the local rank 0 process
+        world = self._rank_world
+        wait_for_local_leader(world)
+
+        # Clean up shared memory resources
+        if hasattr(self, '_next_epoch_shm'):
+            self._cleanup_shared_memory(self._next_epoch_shm, world)
+        if hasattr(self, '_shard_states'):
+            self._cleanup_shared_memory(self._shard_states, world)
+        if hasattr(self, '_resume_shm'):
+            self._cleanup_shared_memory(self._resume_shm, world)
